UNIT ADLIBU;

INTERFACE


TYPE
    WPTR = ^WORD;
    BPTR = ^BYTE;
    BARRAY = ARRAY [0..1] OF BYTE;

{
	 Copyright Ad Lib Inc., 1990

	 This file is part of the Ad Lib Programmer's Manual product and is
	 subject to copyright laws.  As such, you may make copies of this file
	 only for the purpose of having backup copies.  This file may not be
	 redistributed in any form whatsoever.

	 If you find yourself in possession of this file without having received
	 it directly from Ad Lib Inc., then you are in violation of copyright
	 laws, which is a form of theft.
}

{*
	ADLIB.C, low level sound driver V1.04
	===== To be used with OUTCHIP.ASM & SETFREQ.ASM files. =====

	Copyright Ad Lib Inc, 1988, 1989, 1990

        1988/06/22, Marc Savary.
	1989/04/20, Marc savary.
	1990/03/15, Marc Savary.	Modification of frequency calculation (SETFREQ.ASM)
					Correction of volume bugs on additive timbres

        1991/02/22, Bill Parquet!

	The following routines are public (see each routine for further
	documentation):

		SoundColdInit (port)
		SoundWarmInit()
		SetMode (mode)
		SetWaveSel (state)
		SetPitchRange (pR)
		SetGParam (amD, vibD, nSel)
		SetVoiceTimbre (voice, paramArray)
		SetVoiceVolume (voice, volume)
		SetVoicePitch (voice, pitchBend)
		NoteOn (voice, pitch)
		NoteOff (voice)

*}

{$I cflags.inc}
{$I adlib.inc}

{#INCLUDE "ADLIB.H"}

{ Declaring variables as near improves performance. }

{
	In percussive mode, the last 4 voices  (SD TOM HH CYMB) are created
	using melodic voices 7 & 8. A noise generator use channels 7 & 8
	frequency information for creating rhythm instruments. Best results
	are obtained by setting TOM two octaves below mid C and SD 7 half-tones
	above TOM.
	In this implementation, only the TOM pitch may vary, with the SD always
	set 7 half-tones above it.
}

CONST
     TOM_PITCH = 24;				{ best frequency, in range of 0 to 95 }
     TOM_TO_SD = 7;				{ 7 half-tones between voice 7 & 8    }
     SD_PITCH  = TOM_PITCH + TOM_TO_SD;

{#define GetLocPrm(slot, prm) ((unsigned)paramSlot [slot] [prm])}



{
-----------------------------------------------------------------
}

VAR
   pitchRange : INTEGER;
   modeWaveSel : INTEGER;
   percbits : BYTE;

CONST
     percMasks : ARRAY [0..4] OF BYTE = ($10,$08,$04,$02,$01);

VAR
   voiceNote   : ARRAY [0..8]  OF BYTE;  { Pitch of last note-on of each voice }
   voiceKeyOn  : ARRAY [0..8]  OF BYTE;  { State of keyOn bit of each voice }
   vPitchBend  : ARRAY [0..8]  OF WORD;  { Current pitch bend of each voice }
   bxRegister  : ARRAY [0..8]  OF BYTE;  { Current val. of $b0 - $b8 reg    }
   lVoiceVolume: ARRAY [0..10] OF BYTE;  { Volume for each of 11 logical voices }
   modeVoices  : WORD;                   { 9 or 11, depending on 'percussion' }


{ definition of the ELECTRIC-PIANO voice (opr0 & opr1) }

CONST
      pianoParamsOp0 : ARRAY [0..nbLocParam-1] OF BYTE =
                       (1,1,3,15,5,0,1,3,15,0,0,0,1,0);
      pianoParamsOp1 : ARRAY [0..nbLocParam-1] OF BYTE =
	               (0,1,1,15,7,0,2,4,0,0,0,1,0,0);

{ definition of default percussive voices: }

CONST
     bdOpr0 : ARRAY [0..13] OF BYTE = (0,0 ,0,10,4 ,0,8,12,11,0,0,0,1,0);
     bdOpr1 : ARRAY [0..13] OF BYTE = (0,0 ,0,13,4 ,0,6,15, 0,0,0,0,1,0);
     sdOpr  : ARRAY [0..13] OF BYTE = (0,12,0,15,11,0,8, 5, 0,0,0,0,0,0);
     tomOpr : ARRAY [0..13] OF BYTE = (0, 4,0,15,11,0,7, 5, 0,0,0,0,0,0);
     cymbOpr: ARRAY [0..13] OF BYTE = (0, 1,0,15,11,0,5, 5, 0,0,0,0,0,0);
     hhOpr  : ARRAY [0..13] OF BYTE = (0, 1,0,15,11,0,7, 5, 0,0,0,0,0,0);

VAR
   paramSlot : ARRAY [0..17,0..nBLocParam-1] OF BYTE;{ All the parameters of slots }

CONST
     amDepth    : BYTE    = 0;
     vibDepth   : BYTE    = 0;
     noteSel    : BYTE    = 0;
     percussion : BOOLEAN = FALSE;

{Slot numbers as a function of the voice and the operator. (melodic mode only)}

CONST
     slotMVoice : ARRAY [0..8,0..1] OF BYTE =
	((0,3),(1,4),(2,5),(6,9),(7,10),(8,11),(12,15),(13,16),(14,17));

{ Slot numbers for the percussive voices.
   255 indicates that there is only one slot used by a voice. }

CONST
     slotPVoice : ARRAY [0..10,0..1] OF BYTE =
         ((0,3),(1,4),(2,5),(6,9),(7,10),(8,11),(12,15),(16,255),(14,255),(17,255),(13,255));

{
	This table gives the offset of each slot within the chip.
	offset = fn (slot)
}

CONST
     offsetSlot : ARRAY [0..17] OF BYTE =
         (0,1,2,3,4,5,  8,9,10,11,12,13,  16,17,18,19,20,21);

{ This table indicates if the slot is a modulator (operator 0) or a
   carrier (operator 1).
   opr = fn (slot) }

CONST
     carrierSlot : ARRAY [0..17] OF BYTE =
         (0,0,0, 1,1,1, 0,0,0, 1,1,1, 0,0,0, 1,1,1);

{ This table gives the voice number associated with each slot.
   (melodic mode only)
   voice = fn (slot)  }

CONST
     voiceMSlot : ARRAY [0..17] OF BYTE =
         (0,1,2, 0,1,2, 3,4,5, 3,4,5, 6,7,8, 6,7,8);


{ This table gives the voice number  (0-10) associated with each
   slot (percussive mode only),
   voice = fn (slot) }

CONST
     voicePSlot : ARRAY [0..17] OF BYTE =
         (0,1,2, 0,1,2, 3,4,5, 3,4,5, BD,HIHAT,TOM, BD,SD,CYMB);


{----------------------------------------------------------
   Function prototypes.
*/

extern SndOutput();			/* in file OUTCHIP.ASM */
extern SetFreq();				/* in file SETFREQ.ASM */

static InitSlotParams();
static SetSlotParam (unsigned, unsigned *, unsigned);
static SndSetPrm (int, int);
static SndSetAllPrm (int);
static SndSKslLevel (int);
static SndSNoteSel();
static SndSFeedFm (int);
static SndSAttDecay (int);
static SndSSusRelease (int);
static SndSAVEK (int);
static SndSAmVibRhythm();
static SndWaveSelect (int);
static UpdateFNums (int);

FUNCTIONstatic int BoardInstalled();

}

{ Public Functions! }

FUNCTION SoundColdInit : BOOLEAN;
PROCEDURE SoundWarmInit;
PROCEDURE SetMode (mode:BOOLEAN);
PROCEDURE SetWaveSel (state:BOOLEAN);
PROCEDURE SetPitchRange (pR:WORD);
PROCEDURE SetGParam (amD, vibD, nSel:INTEGER);
PROCEDURE SetVoiceTimbre (voice : WORD; paramArray:WPTR);
PROCEDURE SetVoiceVolume (voice, volume:WORD);
PROCEDURE SetVoicePitch (voice, pitchBend:WORD);
PROCEDURE NoteOn (voice : WORD; pitch : INTEGER);
PROCEDURE NoteOff (voice:WORD);
{$F+}
PROCEDURE SndOutput (v1 : INTEGER; v2 : WORD);
FUNCTION  DetectAdlib: INTEGER;
PROCEDURE Messup_Port;
{$F-}

CONST
     fNumTbl : ARRAY [0..191] OF WORD =
	($02b2, $02b4, $02b7, $02b9, $02bc, $02be, $02c1, $02c3, $02c6, $02c9,
	 $02cb, $02ce, $02d0, $02d3, $02d6, $02d8, $02db, $02dd, $02e0, $02e3,
	 $02e5, $02e8, $02eb, $02ed, $02f0, $02f3, $02f6, $02f8, $02fb, $02fe,
	 $0301, $0303, $0306, $0309, $030c, $030f, $0311, $0314, $0317, $031a,
	 $031d, $0320, $0323, $0326, $0329, $032b, $032e, $0331, $0334, $0337,
	 $033a, $033d, $0340, $0343, $0346, $0349, $034c, $034f, $0352, $0356,
	 $0359, $035c, $035f, $0362, $0365, $0368, $036b, $036f, $0372, $0375,
	 $0378, $037b, $037f, $0382, $0385, $0388, $038c, $038f, $0392, $0395,
	 $0399, $039c, $039f, $03a3, $03a6, $03a9, $03ad, $03b0, $03b4, $03b7,
	 $03bb, $03be, $03c1, $03c5, $03c8, $03cc, $03cf, $03d3, $03d7, $03da,
	 $03de, $03e1, $03e5, $03e8, $03ec, $03f0, $03f3, $03f7, $03fb, $03fe,
	 $0fe01, $0fe03, $0fe05, $0fe07, $0fe08, $0fe0a, $0fe0c, $0fe0e, $0fe10, $0fe12,
	 $0fe14, $0fe16, $0fe18, $0fe1a, $0fe1c, $0fe1e, $0fe20, $0fe21, $0fe23, $0fe25,
	 $0fe27, $0fe29, $0fe2b, $0fe2d, $0fe2f, $0fe31, $0fe34, $0fe36, $0fe38, $0fe3a,
	 $0fe3c, $0fe3e, $0fe40, $0fe42, $0fe44, $0fe46, $0fe48, $0fe4a, $0fe4c, $0fe4f,
	 $0fe51, $0fe53, $0fe55, $0fe57, $0fe59, $0fe5c, $0fe5e, $0fe60, $0fe62, $0fe64,
	 $0fe67, $0fe69, $0fe6b, $0fe6d, $0fe6f, $0fe72, $0fe74, $0fe76, $0fe79, $0fe7b,
	 $0fe7d, $0fe7f, $0fe82, $0fe84, $0fe86, $0fe89, $0fe8b, $0fe8d, $0fe90, $0fe92,
	 $0fe95, $0fe97, $0fe99, $0fe9c, $0fe9e, $0fea1, $0fea3, $0fea5, $0fea8, $0feaa,
	 $0fead, $0feaf );

     noteDIV12 : ARRAY [0..95] OF BYTE =
	(0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  1,  1,  1,  1,
	 1,  1,  1,  1,  1,  1,  1,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,
	 2,  2,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  4,  4,  4,
	 4,  4,  4,  4,  4,  4,  4,  4,  4,  5,  5,  5,  5,  5,  5,  5,  5,
	 5,  5,  5,  5,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  7,
	 7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7);

     noteMOD12 : ARRAY [0..95] OF BYTE =
	(0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11,  0,  1,  2,  3,  4,
	 5,  6,  7,  8,  9, 10, 11,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9,
	 10, 11, 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11,  0,  1,  2,
	 3,  4,  5,  6,  7,  8,  9, 10, 11,  0,  1,  2,  3,  4,  5,  6,  7,
	 8,  9, 10, 11,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11,  0,
         1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11);

IMPLEMENTATION


PROCEDURE UpdateFNums (voice: INTEGER); FORWARD;
PROCEDURE InitSlotParams; FORWARD;
PROCEDURE SndSAmVibRhythm; FORWARD;
PROCEDURE SndSNoteSel; FORWARD;
PROCEDURE SetSlotParam (slot : WORD; param : WPTR; waveSel : WORD); FORWARD;
PROCEDURE SndSKslLevel (slot : INTEGER); FORWARD;
PROCEDURE SetCharSlotParam (slot : WORD; cParam : BPTR; waveSel : WORD); FORWARD;
PROCEDURE SndSetPrm (slot, prm:INTEGER); FORWARD;
PROCEDURE SndSetAllPrm (slot : INTEGER); FORWARD;
PROCEDURE SndSFeedFm (slot:INTEGER); FORWARD;
PROCEDURE SndSAttDecay (slot:INTEGER); FORWARD;
PROCEDURE SndSSusRelease (slot:INTEGER); FORWARD;
PROCEDURE SndSAVEK (slot:INTEGER); FORWARD;
PROCEDURE SndWaveSelect (slot: INTEGER); FORWARD;
PROCEDURE SndOutput (v1 : INTEGER; v2 : WORD); EXTERNAL;      {$L OUTCHIP}
PROCEDURE Messup_Port;EXTERNAL;
FUNCTION DetectAdlib: INTEGER; EXTERNAL;
FUNCTION SetFreq (v1, v2, v3, v4: INTEGER) : BYTE; EXTERNAL;  {$L SETFREQ}

{
----------------------------------------------------------
}

{
	Must be called for start-up initialisation.

	Returns 0 if hardware not found.
}

FUNCTION SoundColdInit : BOOLEAN;

BEGIN
     SoundWarmInit;
     SoundColdInit := TRUE;
END;



{
-----------------------------------------------
	Initialize the chip in melodic mode (mode == 0),
	set all 9 voices to electric-piano timbres,
	set the 3 global parameters to zero,
	set the pitch bend range to 1 half-tone,
	set the pitch bend of each voice to 0x2000 (no detune),
	set the volume of each voice to maximum level,
	and enable the wave-select parameter.
-----------------------------------------------
}

PROCEDURE SoundWarmInit;

VAR
   i : BYTE;

BEGIN
     FOR i:=1 TO $f5 DO
         SndOutput (i, 0);      { clear all registers }
     SndOutput (4, 6);          { mask T1 & T2 }

     FOR i:=0 TO 8 DO           { pitch bend for each voice = no detune }
     BEGIN
          vPitchBend [i] := MID_PITCH;
	  voiceKeyOn [i] := 0;
	  voiceNote [i]  := 0;
     END;

     FOR i:=0 TO 10 DO
         lVoiceVolume [i] := MAX_VOLUME;

     SetMode (FALSE);            { melodic mode }
     SetGParam (0, 0, 0);        { init global parameters }
     SetPitchRange (1);          { default pitch range is 1 half-tone }
     SetWaveSel (TRUE);
END;



{
---------------------------------------------
	Put the chip in melodic mode (mode == 0),
	or in percussive mode  (mode != 0).

	If the melodic mode is chosen, all voices are
	set to electric-piano, else the first 5 are set
	to electric-piano, and the percussion voices
	to their default timbres.
---------------------------------------------
}

PROCEDURE SetMode (mode:BOOLEAN);

BEGIN
     IF (mode) THEN
     BEGIN
          { set the frequency for the last 4 percussion voices: }
	  voiceNote [TOM]  := TOM_PITCH;
	  vPitchBend [TOM] := MID_PITCH;
	  UpdateFNums (TOM);

	  voiceNote [SD]   := SD_PITCH;
	  vPitchBend [SD]  := MID_PITCH;
	  UpdateFNums (SD);
     END;

     percussion := mode;
     IF (mode) THEN modeVoices := 11 ELSE modeVoices := 9;
     percBits := 0;

     InitSlotParams;
     SndSAmVibRhythm;
END;

{
	Enable (state != 0) / disable (state == 0)
	the wave-select parameters.

	If you do not want to use the wave-select parameters, call
	this function with a value of 0 AFTER calling SoundColdInit()
	or SoundWarmInit().
}

PROCEDURE SetWaveSel (state:BOOLEAN);

VAR
   i : INTEGER;

BEGIN
     IF (state) THEN modeWaveSel := $20 ELSE modeWaveSel := 0;

     FOR i := 0 TO 17 DO
         SndOutput ($E0 + offsetSlot [i], 0);

     SndOutput (1, modeWaveSel);
END;




{
	Routine to change the pitch bend range. The value can be from
	1 to 12 (in half-tones).

	For example, the value 12 means that the pitch bend will 
	range from -12 (pitchBend == 0, see function 'SetVoicePitch()')
	to +12 (pitchBend == 0x3fff) half-tones.

	The change will be effective as of the next call to
	'SetVoicePitch()'.
}

PROCEDURE SetPitchRange (pR : WORD);

BEGIN
     IF (pR > 12) THEN pR := 12;
     IF (pR < 1)  THEN pR := 1;
     pitchRange := pR;
END;

{
----------------------------------------------
	Set the 3 global parameters AmDepth,
	VibDepth & NoteSel

	The change takes place immediately.
----------------------------------------------
}

PROCEDURE SetGParam (amD, vibD, nSel : INTEGER);

BEGIN
     amDepth  := amD;
     vibDepth := vibD;
     noteSel  := nSel;

     SndSAmVibRhythm;
     SndSNoteSel;
END;




{-------------------------------------------------
	Set the parameters of the voice 'voice'.

	In melodic mode, 'voice' varies from 0 to 8.
	In percussive mode, voices 0 to 5 are melodic
	and 6 to 10 are percussive.

	A timbre (melodic or percussive) is defined as follows:
	the 13 first parameters of operator 0  (ksl, multi, feedBack,
	attack, sustain, eg-typem decay, release, level, am, vib, ksr, fm)
	followed by the 13 parameters of operator 1 (if a percussive voice, all
	the parameters are zero), followed by the wave-select parameter for
	the operators 0 and 1.

	'paramArray' is structured as follows:
		struct
			int opr0Prm [13];       first 13 parameters
			int opr1Prm [13];       must be 0 if percussive timbre
			int	opr0WaveSel;		last parameter
			int opr1WaveSel;		must be 0 if percussive timbre
		 TimbreDef;

	The old timbre files (*.INS) do not contain the parameters
	'opr0WaveSel' and 'opr1WaveSel'.
	Set these two parameters to zero if you are using the old file
	format.
------------------------------------------------
}

PROCEDURE SetVoiceTimbre (voice : WORD; paramArray : WPTR);


VAR
   wave0, wave1 : WORD;
   prm1, wavePtr : WPTR;
   slots : ^BARRAY;

BEGIN
     IF (voice >= modeVoices) THEN EXIT;

     wavePtr := paramArray; INC (wavePtr,2*(nbLocParam-1));
     wave0   := wavePtr^;   INC (wavePtr);
     wave1   := wavePtr^;
     prm1    := paramArray; INC (prm1, nbLocParam -1);

     IF (percussion) THEN
        slots := ADDR (slotPVoice [voice][0])
     ELSE
         slots := ADDR (slotMVoice [voice][0]);

     SetSlotParam (slots^ [0], paramArray, wave0);
     IF (slots^ [1] <> 255) THEN
        SetSlotParam (slots^ [1], prm1, wave1);
END;



{
--------------------------------------------------
	Set the volume of the voice 'voice' to 'volume'.

	The resulting output level is (timbreVolume * volume / 127).
	The change takes place immediately.

	0 <= volume <= 127
--------------------------------------------------
}

PROCEDURE SetVoiceVolume (voice, volume:WORD);

VAR
   slots : ^BARRAY;

BEGIN
     IF (voice >= modeVoices) THEN EXIT;
     IF (volume > MAX_VOLUME) THEN volume := MAX_VOLUME;
     lVoiceVolume [voice] := volume;

     IF (percussion) THEN
        slots := ADDR (slotPVoice [voice][0])
     ELSE
         slots := ADDR (slotMVoice [voice][0]);

     SndSKslLevel (slots^ [0]);
     IF (255 <> slots^ [1]) THEN SndSKslLevel (slots^ [1]);
END;


{
-------------------------------------------------
	Change the pitch value of a voice.

	The variation in pitch is a function of the previous call
	to 'SetPitchRange()' and the value of 'pitchBend'.
	A value of 0 means -half-tone * pitchRange,
	0x2000 means no variation (exact pitch) and
	0x3fff means +half-tone * pitchRange.

	Does not affect the percussive voices, except for the bass drum.

	The change takes place immediately.

	0 <= pitchBend <= 0x3fff, 0x2000 == exact tuning
-------------------------------------------------
}

PROCEDURE SetVoicePitch (voice, pitchBend : WORD);


BEGIN
     IF (((NOT percussion) AND (voice < 9)) OR (voice <= BD)) THEN
     BEGIN
	  { melodic + bass-drum }
	  IF (pitchBend > MAX_PITCH) THEN
	     pitchBend := MAX_PITCH;
	  vPitchBend [voice] := pitchBend;
	  UpdateFNums (voice);
     END;
END;


{*
-----------------------------------------------------------
	Routine to start a note playing.

	0 <= voice <= 8	in melodic mode,
	0 <= voice <= 10 in percussive mode;
	0 <= pitch <= 127, 60 == MID_C  (the card can play between 12 and 107 )
-----------------------------------------------------------
*}

PROCEDURE NoteOn (voice : WORD; pitch : INTEGER);

BEGIN
     DEC (pitch, (MID_C - CHIP_MID_C));
     IF (pitch < 0) THEN pitch := 0;

     IF (((NOT percussion) AND (voice < 9)) OR (voice < BD)) THEN
     BEGIN
	  { this is a melodic voice }
	  voiceNote [voice] := pitch;
	  voiceKeyOn [voice] := $20;
	  UpdateFNums (voice);
     END
     ELSE
	 IF ((percussion) AND (voice <= HIHAT)) THEN
         BEGIN
	      { this is a percussive voice }
	      IF (voice = BD) THEN
              BEGIN
	           voiceNote [BD] := pitch;
		   UpdateFNums (voice);
	      END
              ELSE
		  IF (voice = TOM) THEN
                  BEGIN
			{ for the last 4 percussions, only the TOM may change in frequency,
				modifying the three others: }
			IF (voiceNote [TOM] <> pitch) THEN
                        BEGIN
			     voiceNote [TOM] := pitch;
			     voiceNote [SD]  := pitch +TOM_TO_SD;
			     UpdateFNums (TOM);
			     UpdateFNums (SD);
			END;
		  END;
	      percBits := percBits OR percMasks [voice - BD];
	      SndSAmVibRhythm;
         END;
END;

{
	Routine to stop playing the note which was started in 'NoteOn()'.

	0 <= voice <= 8	in melodic mode,
	0 <= voice <= 10 in percussive mode;
}

PROCEDURE NoteOff (voice : WORD);

BEGIN
     IF (((NOT percussion) AND (voice < 9)) OR (voice < BD)) THEN
     BEGIN
          voiceKeyOn [voice] := 0;
	  bxRegister [voice] := bxRegister [voice] AND (NOT $20);
	  SndOutput ($B0 +voice, bxRegister [voice]);
     END
     ELSE
	 IF ((percussion) AND (voice <= HIHAT)) THEN
         BEGIN
	      percBits := percBits AND (NOT percMasks [voice - BD]);
	      SndSAmVibRhythm;
         END;
END;


{*
------------------------------------------------------------------------
	static functions ...
------------------------------------------------------------------------
*}


{*
	In melodic mode, initialize all voices to electric-pianos.

	In percussive mode, initialize the first 6 voices to electric-pianos
	and the percussive voices to their default timbres.
*}

PROCEDURE InitSlotParams;

VAR
   i : INTEGER;

BEGIN
     FOR i := 0 TO 17 DO
         IF (carrierSlot [i] = 1) THEN SetCharSlotParam (i, ADDR(pianoParamsOp1), 0)
	 ELSE                          SetCharSlotParam (i, ADDR(pianoParamsOp0), 0);

     IF (percussion) THEN
     BEGIN
          SetCharSlotParam (12, ADDR(bdOpr0), 0);
	  SetCharSlotParam (15, ADDR(bdOpr1), 0);
	  SetCharSlotParam (16, ADDR(sdOpr), 0);
	  SetCharSlotParam (14, ADDR(tomOpr), 0);
	  SetCharSlotParam (17, ADDR(cymbOpr), 0);
	  SetCharSlotParam (13, ADDR(hhOpr), 0);
     END;
END;


{
	Used to change the parameter 'param' of the slot 'slot'
	with the value 'val'. The chip registers are updated.
}

PROCEDURE SetASlotParam (slot, param, val : INTEGER);

BEGIN
     paramSlot [slot] [param] := val;
     SndSetPrm (slot, param);
END;

{
------------------------------------------------------
	Set the 14 parameters  (13 in 'param', 1 in 'waveSel')
	of slot 'slot'. Update the parameter array and the chip.
------------------------------------------------------
}

PROCEDURE SetSlotParam (slot : WORD; param : WPTR; waveSel : WORD);

VAR
   i,k : INTEGER;
   ptr : BPTR;

BEGIN
     ptr := ADDR(paramSlot [slot]);

     FOR i := 0 TO nbLocParam-2 DO
     BEGIN
          ptr^ := param^;
          INC (ptr);
          INC (param);
     END;
     waveSel := waveSel AND 3;
     ptr^ := waveSel;
     SndSetAllPrm (slot);
END;

PROCEDURE SetCharSlotParam (slot : WORD; cParam : BPTR; waveSel : WORD);

VAR
   param : ARRAY [0..nbLocParam-1] OF WORD;
   i : INTEGER;

BEGIN
     FOR i := 0 TO nbLocParam -2 DO
     BEGIN
          param [i] := cParam^;
          INC (cParam);
     END;
     SetSlotParam (slot, ADDR(param), waveSel);
END;

{
-----------------------------------------------
	Update the parameter 'prm' for the slot 'slot'.
	Update the chip registers.
-----------------------------------------------
}

PROCEDURE SndSetPrm (slot, prm:INTEGER);

BEGIN
     CASE (prm) OF
          prmPercussion, prmAmDepth, prmVibDepth:
			SndSAmVibRhythm;

	  prmNoteSel:
			SndSNoteSel;

          prmKsl,prmLevel:
			SndSKslLevel (slot);

	  prmFm,prmFeedBack:
			SndSFeedFm (slot);

          prmAttack,prmDecay:
			SndSAttDecay (slot);

	  prmRelease,prmSustain:
			SndSSusRelease (slot);

          prmMulti, prmVib, prmStaining, prmKsr, prmAm:
			SndSAVEK (slot);

	  prmWaveSel:
			SndWaveSelect (slot);

     END;
END;

{-------------------------------------------------
	Transfer all the parameters from slot 'slot' to
	the chip.
}

PROCEDURE SndSetAllPrm (slot : INTEGER);

BEGIN
     SndSAmVibRhythm;
     SndSNoteSel;
     SndSKslLevel (slot);
     SndSFeedFm (slot);
     SndSAttDecay (slot);
     SndSSusRelease (slot);
     SndSAVEK (slot);
     SndWaveSelect (slot);
END;

{
	Write to the register which controls output level and does
	key-on/key-offs for the percussive voice slots.
}

PROCEDURE SndSKslLevel (slot : INTEGER);

VAR
   t1, vc: WORD;
   singleSlot : BOOLEAN;

BEGIN
     IF (percussion) THEN vc := voicePSlot [slot]
     ELSE                 vc := voiceMSlot [slot];

     t1 := 63 - (paramSlot [slot] [prmLevel] AND 63);    { amplitude }
     singleSlot := percussion AND (vc > BD);

     IF (singleSlot OR (carrierSlot [slot]<>0) OR (paramSlot [slot] [prmFm] = 0)) THEN
	{ Change the 0 - 127 volume change value to 0 - 63 for the chip.
	  (MAX_VOLUME+1)/2 is added to avoid round-off errors. }
	t1 := (t1 * WORD(lVoiceVolume [vc]) + (MAX_VOLUME+1) DIV 2) SHR LOG2_VOLUME;
	t1 := 63 - t1;
	t1 := t1 OR paramSlot [slot] [prmKsl] SHL 6;
	SndOutput ($40 + INTEGER(offsetSlot [slot]), t1);
END;



{ --------------------------------------------
	Write to the register which controls the note select parameter.
}

PROCEDURE SndSNoteSel;

BEGIN
     IF (noteSel<>0) THEN SndOutput ($08,64)
     ELSE                 SndOutput ($08,0 );
END;


{ --------------------------------------------
	FEED-BACK and FM (connection).
	Applicable only to operator 0 in melodic mode.
}

PROCEDURE SndSFeedFm (slot:INTEGER);

VAR
   t1 : WORD;

BEGIN
     IF (carrierSlot [slot]<>0) THEN EXIT;
     t1 := paramSlot [slot] [prmFeedBack] SHL 1;
     IF (paramSlot [slot] [prmFm]=0) THEN t1 := t1 OR 1;
     SndOutput ($C0 + INTEGER(voiceMSlot [slot]), t1);
END;


{
	ATTACK, DECAY
}

PROCEDURE SndSAttDecay (slot:INTEGER);

VAR
   t1 : WORD;

BEGIN
     t1 := paramSlot [slot] [prmAttack] SHL 4;
     t1 := t1 OR (paramSlot [slot] [prmDecay] AND $f);
     SndOutput ($60 + INTEGER(offsetSlot [slot]), t1);
END;


{
	SUSTAIN, RELEASE
}

PROCEDURE SndSSusRelease (slot:INTEGER);

VAR
   t1 : WORD;

BEGIN
     t1 := paramSlot [slot] [prmSustain] SHL 4;
     t1 := t1 OR (paramSlot [slot] [prmRelease] AND $f);
     SndOutput ($80 + INTEGER(offsetSlot [slot]), t1);
END;

{
	AM, VIB, EG-TYP (Sustaining), KSR, MULTI
}

PROCEDURE SndSAVEK (slot:INTEGER);

VAR
   t1 : WORD;

BEGIN
     t1 := 0;
     IF (paramSlot [slot] [prmAm] <> 0) THEN t1 := $80;
     IF (paramSlot [slot] [prmVib] <> 0) THEN t1 := t1 OR $40;
     IF (paramSlot [slot] [prmStaining] <> 0) THEN t1 := t1 OR $20;
     IF (paramSlot [slot] [prmKsr] <> 0) THEN t1 := t1 OR $10;
     INC (t1, paramSlot [slot] [prmMulti] AND 15);
     SndOutput ($20 + INTEGER(offsetSlot [slot]), t1);
END;

{
	Set the values: AM Depth, VIB depth & Rhythm (melo/perc mode)
}

PROCEDURE SndSAmVibRhythm;

VAR
   t1 : WORD;

BEGIN
     t1 := 0;
     IF (amDepth <> 0) THEN t1 := $80;
     IF (vibDepth <> 0) THEN t1 := t1 OR $40;
     IF (percussion) THEN t1 := t1 OR $20;
     t1 := t1 OR percBits;
     SndOutput ($BD, t1);
END;


{
	Set the wave-select parameter.
}

PROCEDURE SndWaveSelect (slot: INTEGER);

VAR
   wave : WORD;

BEGIN
     IF (modeWaveSel <> 0) THEN wave := paramSlot [slot] [prmWaveSel] AND $03
     ELSE wave := 0;

     SndOutput ($E0 + offsetSlot [slot], wave);
END;

{
	Change pitch, pitchBend & keyOn of voices 0 to 8, for melodic
	or percussive mode.
}

PROCEDURE UpdateFNums (voice: INTEGER);

BEGIN
     bxRegister [voice] := SetFreq (voice, voiceNote [voice],vPitchBend [voice], voiceKeyOn [voice]);
END;

{
	Return 0 if board is not installed.  The chip's timers are used to
	determined if an Ad Lib card is present.  When being used, the timers
	place specific values in the status register.  If we do not read the
	correct values from the status register, then we can assume that no
	board is present.
}

{----------------------------------------------------------------------}
(*
{#if 0}

{ The following, OutFreq(), is provided as an alternative to SetFreq().
   You may find it easier to understand as SetFreq is written in assembler.
   However, SetFreq is faster so you should definitely use SetFreq. The
   main purpose of including OutFreq() is to show you how to set the
   f-number registers.
}

static unsigned  freqNums [12] = {
/*  C    C#   D    D#   E    F */
   343, 363, 385, 408, 432, 458,
/*  F#   G    G#   A    A#   B */
   485, 514, 544, 577, 611, 647
};

unsigned  OutFreq (voice, pitch, bend, keyOn)
   int voice, pitch, bend, keyOn;
{
   unsigned effNbr, octave, t1;

   /* Integer division by 12 gives the octave (referred to as BLOCK
	  information in the manual). */
   octave = (pitch / 12) - 1;

   /* The remainder of dividing by 12 gives the half-tone within the octave.
	  The freqNums table gives the value to place in the register given the
	  half-tone. */
   effNbr = freqNums [pitch % 12];

   if (bend != 0x2000) {
	  /* Do a pitch bend.
		 pitchRange is the maximum interval for a pitch bend.
		 'bend' is the amount of pitch bend. */
	  long n;

	  /* Below, the maximum interval is calculated in terms of fnums, and
		 then a fraction of this is added to the base value (effNbr). */
	  if (bend > 0x2000) {
		 bend -= 0x2000;
		 n = freqNums [(pitch + pitchRange) % 12];
		 if (n < effNbr) n <<= 1;
		 n = n - effNbr;                        /* interval as f-num */
		 effNbr = effNbr + ((n * bend) >> 13);  /* >> 13 is div by 0x2000 */

		 /* If effNbr exceeds its maximum possible value, bring it into range
			by dividing by two, which lowers the pitch by an octave, and add
			1 to the octave to keep the pitch in the correct octave. */
		 while (effNbr > 1023) {
			effNbr >>= 1;
			octave++;
		 }
	  }
	  else {
		 bend = 0x2000 - bend;
		 n = freqNums [(pitch - pitchRange) % 12];
		 if (n > effNbr) n >>= 1;
		 n = effNbr - n;
		 effNbr = effNbr - ((n * bend) >> 13);

		 /* If effNbr exceeds its minimum desirable value, bring it into range
			by multiplying by two, which raises the pitch by an octave, and
			subtract 1 from the octave to keep the pitch in the correct octave. */
		 while (effNbr < freqNums [0]) {
			effNbr <<= 1;
			octave--;
		 }
	  }
   }

   /* Write the lower 8 bits of the f-number. */
   SndOutput (0xA0 + voice, effNbr);

   /* Arrange the key-on bit, the octave and the upper two bits of the
	  f-number in the correct order and write to the register. */
   t1 = keyOn | (octave << 2) | (effNbr >> 8);

   SndOutput (0xB0 + voice, t1);

   return (t1);
}

#endif

*)

END.
